## -*- coding: utf-8 -*-
################################################################
## Author : Gaël Le Mignot
## Email : gael@pilotsystems.net
################################################################

__author__ = "gael@pilotsystems.net"
__format__ = "plaintext"
__version__ = "$Id: GenericCache.py 19 2009-02-26 15:15:00Z gael.le-mignot $"

import time
import memcache

_marker = []

from threading import RLock

from LRUStock import *
from decorators import *

import string
TRANSLATION = [ '.' for i in range(256) ]
for c in range(33, 127):
    TRANSLATION[c] = chr(c)
TRANSLATION = ''.join(TRANSLATION)

class MemcachedGenericCache(dict):
    """
    This is a cache.
    """
    def __init__(self, server, expiry = None, default_fail = False):
        """
        Constructor
         - Server is the address of the server(s) to use
         - Expiry specify a maximal living time of an item inside the cache
         - If default_fail is True, the default behavior will be to raise
           an exception if not present in the cache
        """
        self.expiry = expiry or 0
        self.default_fail = default_fail

        # Pylibmc client
        self.client = memcache.Client(server)

        # Lock for synchronization
        self.lock = RLock()

    @staticmethod
    def marshall_key(key):
        """
        Marshall a key for python-memcache which is a bit picky
        """
        key = str(key)
        return string.translate(key, TRANSLATION)

    def fetch(self, key, fail = None, onmissing = None):
        """
        Fetch a value
        """
        if fail is None:
            fail = self.default_fail
        
        key = self.marshall_key(key)
        val = self.client.get(key)
        if val is None:
            if fail:
                raise KeyError, key
            return onmissing

        return val

    @synchronized
    def fetch_with_generator(self, key, generator, *args, **kwargs):
        """
        Fetch key, calling generator to build if it doesn't exist
        """
        res = self.fetch(key, onmissing = _marker)
        if res is _marker:
            res = generator(*args, **kwargs)
            self[key] = res
            
        return res
    
    def remove(self, key):
        """
        Expire a value
        """
        key = self.marshall_key(key)
        self.client.delete(key)

    def delete_multi(self, keys):
        """
        Delete multi keys at once
        """
        keys = [ self.marshall_key(key) for key in keys ]
        self.client.delete_multi(keys)

    def reconfigure(self, expiry = None):
        """
        Reconfigure the cache settings
        Warning : do not change the expiry of already stored keys
        """
        self.expiry = expiry

    def insert(self, key, value = _marker):
        """
        Insert an object into the cache

        You can provide a CacheNode class, or directly a key and a value
        """        
        if value is _marker:
            value = key

        key = self.marshall_key(key)
        self.client.set(key, value, time = self.expiry)

    def clear(self):
        """
        Clear the whole cache
        """
        self.client.flush_all()

    def collect(self):
        """
        Collect (remove) all obsoleted items, by default, the removal
        is done lazyly at first access
        Doesn't work on memcached
        """
        pass

    def size(self):
        """
        Get the size of the cache (number of objects stored)
        Doesn't work on memcached
        """
        raise NotImplementedError
    
    def keys(self):
        """
        Get stored keys
        Doesn't work on memcached
        """
        raise NotImplementedError

    def has_key(self, key):
        """
        Do we have such key ?
        """
        return str(key) in self.client

    # Hooks for dictionnary methods
    __getitem__ = fetch
    __setitem__ = insert
    __delitem__ = remove
    __len__ = size
    __contains__ = has_key

